/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.plugin.eclipse;

/*
 * Adapted from org.apache.maven.plugin.eclipse.EclipsePlugin which was released under the following
 * license:
 * <p>
 * Copyright 2001-2005 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language governing permissions
 * and limitations under the License.
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
import org.apache.maven.artifact.resolver.ArtifactCollector;
import org.apache.maven.model.Resource;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.eclipse.BuildCommand;
import org.apache.maven.plugin.eclipse.EclipseSourceDir;
import org.apache.maven.plugin.eclipse.Messages;
import org.apache.maven.plugin.eclipse.writers.EclipseProjectWriter;
import org.apache.maven.plugin.eclipse.writers.EclipseSettingsWriter;
import org.apache.maven.plugin.eclipse.writers.EclipseWriterConfig;
import org.apache.maven.plugin.ide.IdeDependency;
import org.apache.maven.plugin.ide.IdeUtils;
import org.apache.maven.project.MavenProject;
import org.bluestemsoftware.open.eoa.plugin.AbstractDeploymentMojo;
import org.bluestemsoftware.open.eoa.plugin.util.ClasspathHelper;
import org.bluestemsoftware.open.eoa.plugin.util.Constants;
import org.bluestemsoftware.open.eoa.plugin.util.DependencyHelper;
import org.bluestemsoftware.open.eoa.plugin.util.FileFilterImpl;
import org.codehaus.plexus.util.IOUtil;
import org.codehaus.plexus.util.StringUtils;

/**
 * Adapted EclipsePlugin to write classpath derived according to rules defined within EOA
 * specification and to accommodate source and resource sub-directories within it test
 * directory.
 * 
 * @goal eclipse
 * @execute phase="generate-resources"
 */

public class EclipseMojo extends AbstractDeploymentMojo {

    static final String LINE_BREAK = System.getProperty("line.separator");

    /**
     * The currently executed project (can be a reactor project).
     * 
     * @parameter expression="${executedProject}"
     * @readonly
     */
    private MavenProject executedProject;

    /**
     * @component role="org.apache.maven.artifact.metadata.ArtifactMetadataSource" hint="maven"
     */
    private ArtifactMetadataSource metadataSource;

    /**
     * Collects artifacts.
     * 
     * @component
     */
    private ArtifactCollector artifactCollector;

    /**
     * Eclipse workspace directory.
     * 
     * @parameter expression="${eclipse.workspace}" alias="outputDir"
     */
    private File eclipseProjectDir;

    /**
     * Enables/disables the downloading of source attachments. Defaults to false. When this
     * flag is <code>true</code> remote repositories are checked for sources: in order to
     * avoid repeated check for unavailable source archives, a status cache is mantained into
     * the target dir of the root project. Run <code>mvn:clean</code> or delete the file
     * <code>mvn-eclipse-cache.properties</code> in order to reset this cache.
     * @parameter expression="${downloadSources}"
     */
    private boolean downloadSources;

    /**
     * Enables/disables the downloading of javadoc attachments. Defaults to false. When this
     * flag is <code>true</code> remote repositories are checked for javadocs: in order to
     * avoid repeated check for unavailable javadoc archives, a status cache is mantained into
     * the target dir of the root project. Run <code>mvn:clean</code> or delete the file
     * <code>mvn-eclipse-cache.properties</code> in order to reset this cache.
     * 
     * @parameter expression="${downloadJavadocs}"
     */
    private boolean downloadJavadocs;

    /**
     * List of eclipse project natures. By default the
     * <code>org.eclipse.jdt.core.javanature</code> nature plus the needed WTP natures are
     * added. Natures added using this property <strong>replace</strong> the default list.
     * 
     * <pre>
     *                                &lt;projectnatures&gt;
     *                                   &lt;projectnature&gt;org.eclipse.jdt.core.javanature&lt;/projectnature&gt;
     *                                   &lt;projectnature&gt;org.eclipse.wst.common.modulecore.ModuleCoreNature&lt;/projectnature&gt;
     *                                &lt;/projectnatures&gt;
     * </pre>
     * 
     * @parameter
     */
    private List<String> projectnatures;

    /**
     * List of eclipse project natures to be added to the default ones.
     * 
     * <pre>
     *                                &lt;additionalProjectnatures&gt;
     *                                   &lt;projectnature&gt;org.springframework.ide.eclipse.core.springnature&lt;/projectnature&gt;
     *                                &lt;/additionalProjectnatures&gt;
     * </pre>
     * 
     * @parameter
     */
    private List<String> additionalProjectnatures;

    /**
     * List of eclipse project facets to be added to the default ones.
     * 
     * <pre>
     *                                &lt;additionalProjectFacets&gt;
     *                                   &lt;jst.jsf&gt;1.1&lt;jst.jsf/&gt;
     *                                &lt;/additionalProjectFacets&gt;
     * </pre>
     * 
     * @parameter
     */
    private Map<?, ?> additionalProjectFacets;

    /**
     * List of eclipse build commands. By default the
     * <code>org.eclipse.jdt.core.javabuilder</code> builder plus the needed WTP builders are
     * added.
     * 
     * If you specify any configuration for this parameter, only those buildcommands specified
     * will be used; the defaults won't be added. Use the <code>additionalBuildCommands</code>
     * parameter for that.
     * 
     * Configuration example:
     * 
     * Old style:
     * 
     * <pre>
     *                                &lt;buildcommands&gt;
     *                                   &lt;buildcommand&gt;org.eclipse.wst.common.modulecore.ComponentStructuralBuilder&lt;/buildcommand&gt;
     *                                   &lt;buildcommand&gt;org.eclipse.jdt.core.javabuilder&lt;/buildcommand&gt;
     *                                   &lt;buildcommand&gt;org.eclipse.wst.common.modulecore.ComponentStructuralBuilderDependencyResolver&lt;/buildcommand&gt;
     *                                &lt;/buildcommands&gt;
     * </pre>
     * 
     * For new style, see <code>additionalBuildCommands</code>.
     * 
     * @parameter
     */
    private List<BuildCommand> buildcommands;

    /**
     * List of eclipse build commands to be added to the default ones.
     * 
     * Old style:
     * 
     * <pre>
     *                                &lt;additionalBuildcommands&gt;
     *                                   &lt;buildcommand&gt;org.springframework.ide.eclipse.core.springbuilder&lt;/buildcommand&gt;
     *                                &lt;/additionalBuildcommands&gt;
     * </pre>
     * 
     * New style:
     * 
     * <pre>
     *                                &lt;additionalBuildcommands&gt;
     *                                   &lt;buildCommand&gt;
     *                                     &lt;name&gt;org.ui.externaltools.ExternalToolBuilder&lt;/name&gt;
     *                                     &lt;triggers&gt;auto,full,incremental,&lt;/triggers&gt;
     *                                     &lt;arguments&gt;
     *                                       &lt;LaunchConfigHandle&gt;&amp;lt;project&amp;gt;./externalToolBuilders/MavenBuilder.launch&lt;/LaunchConfighandle&gt;
     *                                     &lt;/arguments&gt;
     *                                   &lt;/buildCommand&gt;
     *                                &lt;/additionalBuildcommands&gt;
     * </pre>
     * 
     * Note the difference between <code>build<strong>c</strong>ommand</code> and
     * <code>build<strong>C</strong>ommand</code>.
     * 
     * You can mix and match old and new-style configuration entries.
     * 
     * @parameter
     */
    private List<BuildCommand> additionalBuildcommands;

    /**
     * List of container classpath entries. By default the
     * <code>org.eclipse.jdt.launching.JRE_CONTAINER</code> classpath container is added.
     * Configuration example:
     * 
     * <pre>
     *                                &lt;classpathContainers&gt;
     *                                   &lt;classpathContainer&gt;org.eclipse.jdt.launching.JRE_CONTAINER&lt;/classpathContainer&gt;
     *                                   &lt;classpathContainer&gt;org.eclipse.jst.server.core.container/org.eclipse.jst.server.tomcat.runtimeTarget/Apache Tomcat v5.5&lt;/classpathContainer&gt;
     *                                   &lt;classpathContainer&gt;org.eclipse.jst.j2ee.internal.web.container/artifact&lt;/classpathContainer&gt;
     *                                &lt;/classpathContainers&gt;
     * </pre>
     * 
     * @parameter
     */
    private List<String> classpathContainers;

    /**
     * The default output directory
     * 
     * @parameter expression="${outputDirectory}" alias="outputDirectory"
     *            default-value="${project.build.outputDirectory}"
     * @required
     */
    private File buildOutputDirectory;

    /**
     * The relative path of the manifest file
     * 
     * @parameter expression="${eclipse.manifest}"
     *            default-value="${basedir}/META-INF/MANIFEST.MF"
     */
    private File manifest;

    private boolean isEOAProject = true;    
    
    /**
     * @see org.apache.maven.plugin.Mojo#execute()
     */
    public final void execute() throws MojoExecutionException, MojoFailureException {

        boolean processProject = setup();
        if (!processProject) {
            return;
        }

        IdeDependency[] deps = doDependencyResolution();

        resolveSourceAndJavadocArtifacts(deps);

        writeConfiguration(deps);

        reportMissingArtifacts();

    }

    private void writeConfiguration(IdeDependency[] deps) throws MojoExecutionException {

        EclipseWriterConfig config = createEclipseWriterConfig(deps);

        new EclipseSettingsWriter().init(getLog(), config).write();

        new ClasspathWriter().init(getLog(), config).write();

        // NOTE: This one MUST be after EclipseClasspathwriter, and possibly others,
        // since currently EclipseClasspathWriter does some magic to detect nested
        // output folders and modifies the configuration by adding new (Ant) builders.
        // So the .project file must be written AFTER those have run!
        new EclipseProjectWriter().init(getLog(), config).write();

        getLog().info(
                Messages.getString("EclipsePlugin.wrote", new Object[] { config.getEclipseProjectName(),
                        eclipseProjectDir.getAbsolutePath() }));
    }

    /**
     * @see org.apache.maven.plugin.Mojo#execute()
     */
    private boolean setup() throws MojoExecutionException {

        boolean ready = true;

        ready = validate();

        String packaging = executedProject.getPackaging();
        
        // the maven eclipse plugin does not exclude .svn from source classpath entries
        // as of 7/2008. so ... we will process all project types in addition to eoa
        // projects

        boolean skip = false;
        if (!packaging.equals("eoa-component")) {
            if (!packaging.equals("eoa-factory")) {
                if (!packaging.equals("eoa-feature")) {
                    if (!packaging.equals("war") && !packaging.equals("jar") && !packaging.equals("maven-plugin")) {
                        skip = true;
                    } else {
                        isEOAProject = false;
                    }
                }
            }
        }
        
        if (skip) {
            getLog().info("skipping non supported project type '" + packaging + "'.");
            return false;
        }

        if (projectnatures == null) {
            fillDefaultNatures(packaging);
        }

        if (additionalProjectnatures != null) {
            projectnatures.addAll(additionalProjectnatures);
        }

        if (buildcommands == null) {
            fillDefaultBuilders(packaging);
        } else {
            convertBuildCommandList(buildcommands);
        }

        if (additionalBuildcommands != null) {
            convertBuildCommandList(additionalBuildcommands);
            buildcommands.addAll(additionalBuildcommands);
        }

        if (classpathContainers == null) {
            fillDefaultClasspathContainers(packaging);
        } else {
            verifyClasspathContainerListIsComplete();
        }

        return ready;

    }

    private EclipseSourceDir[] buildDirectoryList(MavenProject project, File buildOutputDirectory) throws MojoExecutionException {

        File projectBaseDir = project.getFile().getParentFile();

        Set<EclipseSourceDir> directories = new TreeSet<EclipseSourceDir>();

        // add project's main source and resource directory

        String relativeOutput = IdeUtils.toRelativeAndFixSeparator(projectBaseDir, buildOutputDirectory, false);

        extractSourceDirs(directories, project.getCompileSourceRoots(), projectBaseDir, false, relativeOutput);

        extractResourceDirs(directories, project.getBuild().getResources(), project, projectBaseDir, false,
                relativeOutput);

        // add project's unit test source and resource directory

        String relativeTestOutput = IdeUtils.toRelativeAndFixSeparator(projectBaseDir, new File(project.getBuild()
                .getTestOutputDirectory()), false);

        extractSourceDirs(directories, project.getTestCompileSourceRoots(), projectBaseDir, true,
                relativeTestOutput);

        extractResourceDirs(directories, project.getBuild().getTestResources(), project, projectBaseDir, true,
                relativeTestOutput);

        // add project's integration test source and resource directories

        File it = new File(basedir, "src/it/");

        if (it.exists() && it.isDirectory()) {

            File[] suites = it.listFiles(new FileFilterImpl());
            for (int i = 0; i < suites.length; i++) {

                File suite = suites[i];
                if (!suite.isDirectory()) {
                    continue;
                }

                File system = new File(suite, "system/");
                if (system.exists() && system.isDirectory()) {
                    List<?> sourceRoots = Arrays.asList(new File(system, "java/").getAbsolutePath());
                    relativeTestOutput = "target/it/" + suite.getName() + "/system/test-classes/";
                    extractSourceDirs(directories, sourceRoots, projectBaseDir, true, relativeTestOutput);
                    File testResourcesDir = new File(system, "resources/");
                    if (testResourcesDir.exists() && testResourcesDir.isDirectory()) {
                        Resource resource = new Resource();
                        resource.setDirectory(testResourcesDir.getAbsolutePath());
                        List<?> testResources = Arrays.asList(resource);
                        extractResourceDirs(directories, testResources, project, projectBaseDir, true,
                                relativeTestOutput);
                    }
                }

                File client = new File(suite, "client/");
                if (client.exists() && client.isDirectory()) {
                    List<?> sourceRoots = Arrays.asList(new File(client, "java/").getAbsolutePath());
                    relativeTestOutput = "target/it/" + suite.getName() + "/client/test-classes/";
                    extractSourceDirs(directories, sourceRoots, projectBaseDir, true, relativeTestOutput);
                    File testResourcesDir = new File(client, "resources/");
                    if (testResourcesDir.exists() && testResourcesDir.isDirectory()) {
                        Resource resource = new Resource();
                        resource.setDirectory(testResourcesDir.getAbsolutePath());
                        List<?> testResources = Arrays.asList(resource);
                        extractResourceDirs(directories, testResources, project, projectBaseDir, true,
                                relativeTestOutput);
                    }
                }

            }
        }

        return directories.toArray(new EclipseSourceDir[directories.size()]);
    }

    private void extractSourceDirs(Set<EclipseSourceDir> directories, List<?> sourceRoots, File projectBaseDir, boolean test, String output) throws MojoExecutionException {

        for (Iterator<?> it = sourceRoots.iterator(); it.hasNext();) {

            File sourceRootFile = new File((String)it.next());

            if (!sourceRootFile.isDirectory()) {
                continue;
            }

            String sourceRoot = IdeUtils.toRelativeAndFixSeparator(projectBaseDir, sourceRootFile, !projectBaseDir
                    .equals(basedir));

            directories.add(new EclipseSourceDir(sourceRoot, output, false, false, null, "**/.svn/", false));

        }
    }

    private void extractResourceDirs(Set<EclipseSourceDir> directories, List<?> resources, MavenProject project, File workspaceProjectBaseDir, boolean test, final String output) throws MojoExecutionException {
        for (Iterator<?> it = resources.iterator(); it.hasNext();) {
            Resource resource = (Resource)it.next();

            getLog().debug("Processing resource dir: " + resource.getDirectory());

            String includePattern = null;
            String excludePattern = null;

            if (resource.getIncludes().size() != 0) {
                includePattern = StringUtils.join(resource.getIncludes().iterator(), "|");
            }

            if (resource.getExcludes().size() != 0) {
                excludePattern = StringUtils.join(resource.getExcludes().iterator(), "|");
            }
            
            // cvs dirs and files are excluded by default. not svn files, however
            // so hard code this exclusion pattern
            
            if (excludePattern == null) {
                excludePattern = "**/.svn/";
            } else {
                excludePattern = excludePattern + " | **/.svn/";
            }

            // TODO: figure out how to merge if the same dir is specified twice
            // with different in/exclude patterns.

            File resourceDirectory = new File( /* basedir, */resource.getDirectory());

            if (!resourceDirectory.exists() || !resourceDirectory.isDirectory()) {
                getLog().debug("Resource dir: " + resourceDirectory + " either missing or not a directory.");
                continue;
            }

            String resourceDir = IdeUtils.toRelativeAndFixSeparator(workspaceProjectBaseDir, resourceDirectory,
                    !workspaceProjectBaseDir.equals(basedir));
            String thisOutput = output;
            if (thisOutput != null) {
                // sometimes thisOutput is already an absolute path
                File outputFile = new File(thisOutput);
                if (!outputFile.isAbsolute()) {
                    outputFile = new File(workspaceProjectBaseDir, thisOutput);
                }
                // create output dir if it doesn't exist
                outputFile.mkdirs();

                if (!StringUtils.isEmpty(resource.getTargetPath())) {
                    outputFile = new File(outputFile, resource.getTargetPath());
                    // create output dir if it doesn't exist
                    outputFile.mkdirs();
                }

                getLog().debug(
                        "Making relative and fixing separator: { "
                                + workspaceProjectBaseDir
                                + ", "
                                + outputFile
                                + ", false }.");
                thisOutput = IdeUtils.toRelativeAndFixSeparator(workspaceProjectBaseDir, outputFile, false);
            }

            getLog().debug(
                    "Adding eclipse source dir: { "
                            + resourceDir
                            + ", "
                            + thisOutput
                            + ", true, "
                            + test
                            + ", "
                            + includePattern
                            + ", "
                            + excludePattern
                            + " }.");

            directories.add(new EclipseSourceDir(resourceDir, thisOutput, true, test, includePattern,
                    excludePattern, resource.isFiltering()));
        }
    }

    /**
     * Not a plugin parameter. Collect the list of dependencies with a missing source artifact
     * for the final report.
     */
    private List<IdeDependency> missingSourceDependencies = new ArrayList<IdeDependency>();

    /**
     * Not a plugin parameter. Collect the list of dependencies with a missing javadoc artifact
     * for the final report.
     */
    // TODO merge this with the missingSourceDependencies in a classifier based map?
    private List<IdeDependency> missingJavadocDependencies = new ArrayList<IdeDependency>();

    private IdeDependency[] doDependencyResolution() throws MojoExecutionException {

        if (executedProject.getDependencies() == null) {
            return new IdeDependency[0];
        }

        ClasspathHelper cph = new ClasspathHelper(logger, basedir, executedProject, false);
        cph.setArtifactCollector(artifactCollector);
        cph.setArtifactFactory(artifactFactory);
        cph.setArtifactResolver(artifactResolver);
        cph.setLocalRepository(localRepository);
        cph.setMetadataSource(metadataSource);
        cph.setRemoteRepositories(remoteRepositories);
        cph.setTest(test);

        Set<Artifact> deps = DependencyHelper.getProjectDependencies(executedProject, artifactFactory);

        Set<String> versionlessKeys = new HashSet<String>();
        Set<Artifact> artifacts = new HashSet<Artifact>();
        
        if (isEOAProject) {
        
            artifacts.addAll(cph.getParentClasspathArtifacts(deps, versionlessKeys));
            artifacts.addAll(cph.getSharedClasspathArtifacts(deps, versionlessKeys, true));
            artifacts.addAll(cph.getDeploymentClasspathArtifacts(deps, versionlessKeys));
            artifacts.addAll(cph.getSystemTestClasspathArtifacts(deps, versionlessKeys));
            artifacts.addAll(cph.getClientTestClasspathArtifacts(deps, versionlessKeys));
    
            if (logger.isDebugEnabled()) {
                logger.debug(LINE_BREAK);
                logger.debug("classpath: ");
                for (Artifact a : artifacts) {
                    String c = a.getFile().getName();
                    String dt = a.getDependencyTrail() == null ? "root" : a.getDependencyTrail().toString();
                    logger.debug("constituent: " + c + " dependency trail: " + dt);
                }
                logger.debug("end classpath");
            }
        
        } else {
            
            artifacts.addAll(cph.getNonEOAProjectArtifacts(deps));
            
        }

        List<IdeDependency> dependencies = new ArrayList<IdeDependency>();

        for (Artifact a : artifacts) {
            IdeDependency dep = new IdeDependency(a.getGroupId(), a.getArtifactId(), a.getVersion(), a
                    .getClassifier(), false, Artifact.SCOPE_TEST.equals(a.getScope()), Artifact.SCOPE_SYSTEM
                    .equals(a.getScope()), Artifact.SCOPE_PROVIDED.equals(a.getScope()), a.getArtifactHandler()
                    .isAddedToClasspath(), a.getFile(), a.getType(), false, null, 0);
            dependencies.add(dep);
        }

        // @todo a final report with the list of
        // missingArtifacts?

        return dependencies.toArray(new IdeDependency[dependencies.size()]);

    }

    /**
     * Find the reactor target dir. executedProject doesn't have the multiproject root dir set,
     * and the only way to extract it is iterating on parent projects.
     * @param prj
     *        current project
     * @return the parent target dir.
     */
    private File getReactorTargetDir(MavenProject prj) {
        if (prj.getParent() != null) {
            if (prj.getParent().getBasedir() != null && prj.getParent().getBasedir().exists()) {
                return getReactorTargetDir(prj.getParent());
            }
        }
        return new File(prj.getBuild().getDirectory());
    }

    /**
     * Resolve source artifacts and download them if <code>downloadSources</code> is
     * <code>true</code>. Source and javadocs artifacts will be attached to the
     * <code>IdeDependency</code> Resolve source and javadoc artifacts. The resolved
     * artifacts will be downloaded based on the <code>downloadSources</code> and
     * <code>downloadJavadocs</code> attributes. Source and
     * @param deps
     *        resolved dependencies
     */
    private void resolveSourceAndJavadocArtifacts(IdeDependency[] deps) {

        File reactorTargetDir = getReactorTargetDir(executedProject);
        File unavailableArtifactsTmpFile = new File(reactorTargetDir, "mvn-eclipse-cache.properties");

        getLog().info("Using source status cache: " + unavailableArtifactsTmpFile.getAbsolutePath());

        // create target dir if missing
        if (!unavailableArtifactsTmpFile.getParentFile().exists()) {
            unavailableArtifactsTmpFile.getParentFile().mkdirs();
        }

        Properties unavailableArtifactsCache = new Properties();
        if (unavailableArtifactsTmpFile.exists()) {
            InputStream is = null;
            try {
                is = new FileInputStream(unavailableArtifactsTmpFile);
                unavailableArtifactsCache.load(is);
            } catch (IOException e) {
                getLog().warn("Unable to read source status for reactor projects");
            } finally {
                IOUtil.close(is);
            }

        }

        final List<IdeDependency> missingSources = resolveDependenciesWithClassifier(deps, "sources",
                downloadSources, unavailableArtifactsCache);
        missingSourceDependencies.addAll(missingSources);

        final List<IdeDependency> missingJavadocs = resolveDependenciesWithClassifier(deps, "javadoc",
                downloadJavadocs, unavailableArtifactsCache);
        missingJavadocDependencies.addAll(missingJavadocs);

        FileOutputStream fos = null;
        try {
            fos = new FileOutputStream(unavailableArtifactsTmpFile);
            unavailableArtifactsCache.store(fos, "Temporary index for unavailable sources and javadocs");
        } catch (IOException e) {
            getLog().warn("Unable to cache source status for reactor projects");
        } finally {
            IOUtil.close(fos);
        }

    }

    /**
     * Resolve the required artifacts for each of the dependency. <code>sources</code> or
     * <code>javadoc</code> artifacts (depending on the <code>classifier</code>) are
     * attached to the dependency.
     * 
     * @param deps
     *        resolved dependencies
     * @param classifier
     *        the classifier we are looking for (either <code>sources</code> or
     *        <code>javadoc</code>)
     * @param includeRemoteRepositories
     *        flag whether we should search remote repositories for the artifacts or not
     * @param unavailableArtifactsCache
     *        cache of unavailable artifacts
     * @return the list of dependencies for which the required artifact was not found
     */
    private List<IdeDependency> resolveDependenciesWithClassifier(IdeDependency[] deps, String classifier, boolean includeRemoteRepositories, Properties unavailableArtifactsCache) {
        List<IdeDependency> missingClassifierDependencies = new ArrayList<IdeDependency>();

        // if downloadSources is off, just check
        // local repository for reporting missing source jars
        List<?> remoteRepos = includeRemoteRepositories ? remoteRepositories : Collections.EMPTY_LIST;

        for (int j = 0; j < deps.length; j++) {
            IdeDependency dependency = deps[j];

            if (dependency.isReferencedProject()) {
                // artifact not needed
                continue;
            }
            
            if (getLog().isDebugEnabled()) {
                getLog().debug(
                        "Searching for attachments for "
                                + dependency.getId()
                                + ":"
                                + dependency.getClassifier()
                                + " at "
                                + dependency.getId()
                                + ":"
                                + classifier);
            }

            if (dependency.isSystemScoped()) {
                String artifactName = dependency.getFile().getName();
                int index = artifactName.lastIndexOf(".");
                String attachmentName = artifactName.substring(0, index);
                attachmentName = attachmentName + "-" + classifier + ".jar";
                File attachmentFile = new File(dependency.getFile().getParentFile(), attachmentName);
                if (attachmentFile.exists()) {
                    if (classifier.equals("sources")) {
                        dependency.setSourceAttachment(attachmentFile);
                    } else if (classifier.equals("javadoc")){
                        dependency.setJavadocAttachment(attachmentFile);
                    } else {
                        missingClassifierDependencies.add(dependency);
                    }
                } else {
                    missingClassifierDependencies.add(dependency);
                }
                continue;
            }

            if (!unavailableArtifactsCache.containsKey(dependency.getId() + ":" + classifier)) {
                Artifact artifact = IdeUtils.resolveArtifactWithClassifier(dependency.getGroupId(), dependency
                        .getArtifactId(), dependency.getVersion(), classifier, localRepository, artifactResolver,
                        artifactFactory, remoteRepos, getLog());
                if (artifact.isResolved()) {
                    if ("sources".equals(classifier)) {
                        dependency.setSourceAttachment(artifact.getFile());
                    } else if ("javadoc".equals(classifier)) {
                        dependency.setJavadocAttachment(artifact.getFile());
                    } else {
                        unavailableArtifactsCache.put(dependency.getId() + ":" + classifier, Boolean.TRUE
                                .toString());
                        // add the dependencies to the list
                        // of those lacking the required
                        // artifact
                        missingClassifierDependencies.add(dependency);
                    }
                }

            }
        }

        // return the list of dependencies missing the required artifact
        return missingClassifierDependencies;

    }

    /**
     * Output a message with the list of missing dependencies and info on how turn download on
     * if it was disabled.
     */
    private void reportMissingArtifacts() {
        StringBuffer msg = new StringBuffer();

        if (!missingSourceDependencies.isEmpty()) {
            if (downloadSources) {
                msg.append(Messages.getString("sourcesnotavailable"));
            } else {
                msg.append(Messages.getString("sourcesnotdownloaded"));
            }

            for (Iterator<IdeDependency> it = missingSourceDependencies.iterator(); it.hasNext();) {
                IdeDependency art = it.next();
                msg.append(Messages.getString("sourcesmissingitem", art.getId()));
            }
            msg.append("\n");
        }

        if (!missingJavadocDependencies.isEmpty()) {
            if (downloadJavadocs) {
                msg.append(Messages.getString("javadocnotavailable"));
            } else {
                msg.append(Messages.getString("javadocnotdownloaded"));
            }

            for (Iterator<IdeDependency> it = missingJavadocDependencies.iterator(); it.hasNext();) {
                IdeDependency art = it.next();
                msg.append(Messages.getString("javadocmissingitem", art.getId()));
            }
            msg.append("\n");
        }
        getLog().info(msg);
    }

    private void convertBuildCommandList(List<BuildCommand> commands) {
        if (commands != null) {
            List<BuildCommand> temp = new ArrayList<BuildCommand>();
            for (ListIterator<BuildCommand> i = commands.listIterator(); i.hasNext();) {
                Object command = i.next();
                if (command instanceof String) {
                    temp.add(new BuildCommand((String)command));
                } else {
                    temp.add((BuildCommand)command);
                }
            }
            commands = temp;
        }
    }

    private void verifyClasspathContainerListIsComplete() {
        // this is an extension point.
        if (!classpathContainers.contains(Constants.COMMON_PATH_JDT_LAUNCHING_JRE_CONTAINER)) {
            getLog().warn(Messages.getString("EclipsePlugin.missingjrecontainer"));
            classpathContainers.add(0, Constants.COMMON_PATH_JDT_LAUNCHING_JRE_CONTAINER);
        }
    }

    private boolean validate() throws MojoExecutionException {
        // validate sanity of the current m2 project

        String packaging = executedProject.getPackaging();

        assertNotEmpty(executedProject.getGroupId(), Constants.POM_ELT_GROUP_ID);
        assertNotEmpty(executedProject.getArtifactId(), Constants.POM_ELT_ARTIFACT_ID);

        if (executedProject.getFile() == null || !executedProject.getFile().exists()) {
            throw new MojoExecutionException(Messages.getString("EclipsePlugin.missingpom"));
        }

        if ("pom".equals(packaging) && eclipseProjectDir == null) {
            getLog().info(Messages.getString("EclipsePlugin.pompackaging"));
            return false;
        }

        if (eclipseProjectDir == null) {
            eclipseProjectDir = executedProject.getFile().getParentFile();
        }

        if (!eclipseProjectDir.exists() && !eclipseProjectDir.mkdirs()) {
            throw new MojoExecutionException(Messages.getString("EclipsePlugin.cantcreatedir", eclipseProjectDir));
        }

        if (!eclipseProjectDir.equals(executedProject.getFile().getParentFile())) {
            if (!eclipseProjectDir.isDirectory()) {
                throw new MojoExecutionException(Messages.getString("EclipsePlugin.notadir", eclipseProjectDir));
            }
            eclipseProjectDir = new File(eclipseProjectDir, executedProject.getArtifactId());
            if (!eclipseProjectDir.isDirectory() && !eclipseProjectDir.mkdirs()) {
                throw new MojoExecutionException(Messages.getString("EclipsePlugin.cantcreatedir",
                        eclipseProjectDir));
            }
        }

        return true;
    }

    private EclipseWriterConfig createEclipseWriterConfig(IdeDependency[] deps) throws MojoExecutionException {
        File projectBaseDir = executedProject.getFile().getParentFile();

        // build a list of UNIQUE source dirs (both src and resources) to be
        // used in classpath and wtpmodules
        EclipseSourceDir[] sourceDirs = buildDirectoryList(executedProject, buildOutputDirectory);

        EclipseWriterConfig config = new EclipseWriterConfig();

        config.setProjectNameTemplate(IdeUtils.PROJECT_NAME_DEFAULT_TEMPLATE);

        String projectName = IdeUtils.getProjectName(config.getProjectNameTemplate(), executedProject);

        config.setEclipseProjectName(projectName);

        Set<BuildCommand> convertedBuildCommands = new LinkedHashSet<BuildCommand>();

        if (buildcommands != null) {
            for (Iterator<BuildCommand> it = buildcommands.iterator(); it.hasNext();) {
                Object cmd = it.next();

                if (cmd instanceof BuildCommand) {
                    convertedBuildCommands.add((BuildCommand)cmd);
                } else {
                    convertedBuildCommands.add(new BuildCommand((String)cmd));
                }
            }
        }

        config.setBuildCommands(new LinkedList<BuildCommand>(convertedBuildCommands));

        config.setBuildOutputDirectory(buildOutputDirectory);
        config.setClasspathContainers(classpathContainers);
        config.setDeps(deps);
        config.setEclipseProjectDirectory(eclipseProjectDir);
        config.setLocalRepository(localRepository);
        config.setManifestFile(manifest);
        config.setPde(false);
        config.setProject(executedProject);
        config.setProjectBaseDir(projectBaseDir);
        config.setProjectnatures(projectnatures);
        config.setProjectFacets(additionalProjectFacets);
        config.setSourceDirs(sourceDirs);

        return config;
    }

    private void assertNotEmpty(String string, String elementName) throws MojoExecutionException {
        if (string == null) {
            throw new MojoExecutionException(Messages.getString("EclipsePlugin.missingelement", elementName));
        }
    }

    private void fillDefaultNatures(String packaging) {
        projectnatures = new ArrayList<String>();
        projectnatures.add(Constants.NATURE_JDT_CORE_JAVA);
    }

    private void fillDefaultClasspathContainers(String packaging) {
        classpathContainers = new ArrayList<String>();
        classpathContainers.add(Constants.COMMON_PATH_JDT_LAUNCHING_JRE_CONTAINER);
    }

    private void fillDefaultBuilders(String packaging) {
        buildcommands = new ArrayList<BuildCommand>();
        buildcommands.add(new BuildCommand(Constants.BUILDER_JDT_CORE_JAVA));
    }

}
